const express = require('express');
const router = express.Router();

const fse = require('fs-extra');
const crypto = require('crypto');
const _ = require('lodash');

const axios = require('axios');

const MAX_CODE_SUBMISSIONS = 5;

const {
    checkCompensated,
    checkCompleted,
    createUser,
    getUserData,
    markCompleted,
    recordCompensation,
    updateUser,
} = require('./utilities/utils');

router.post('/signup', (req, res) => {
    const {
        userType,
        sessionId
    } = req.body;

    const userData = createUser(sessionId, userType);
    return res.send({ success: true });
});

router.post('/login', (req, res) => {
    const {
        userType,
        sessionId
    } = req.body;

    const userData = getUserData(sessionId);
    userData.userType = userType;
    const success = updateUser(userData);

    return res.send({ success: !!success });
});

router.post('/userStatus', (req, res) => {
    const { sessionId } = req.body;
    const userData = getUserData(sessionId);

    if (userData.completed) {
        return res.send({ success: true, section: 'completed' });
    } else if (userData.userType !== 'demo' && checkCompleted(sessionId, req.headers['x-forwarded-for'])) {
        return res.send({ success: true, section: 'completed_illegitimate' });
    }

    let { sectionIdx, snippetIdx, questionIdx, sections, snippets, responses } = userData;
    const relevantSection = sections[sectionIdx]; // comprehension, methodWrite, repair
    const relevantSnippets = snippets[relevantSection];
    const question = relevantSnippets[snippetIdx].questions[questionIdx];

    const fid = relevantSnippets[snippetIdx].fid;
    const relevantResponses = responses[relevantSection];
    const response = relevantResponses[fid];

    if (!response) {
        responses[relevantSection][fid] = {
            start: Date.now(),
            questions: [{ question: question.data, start: Date.now() }]
        }
        updateUser(userData);
    }
    
    return res.send({
        success: true,
        data: relevantSnippets[snippetIdx],
        section: relevantSection,
        snippetIdx,
        question,
        response
    });
});

router.post('/completeSurvey', (req, res) => {
    const { sessionId, comments, email } = req.body;
    const userData = getUserData(sessionId);

    // If demoing, don't save that the user has completed the survey
    if (userData.userType == 'demo') {
        return res.send({ success: true, type: 'demo' });
    }

    const isCompleted = checkCompleted(sessionId);
    if (!isCompleted) {
        return res.send({ success: false, err: 'User has not completed the survey' });
    }
    const isCompensated = checkCompensated(sessionId);
    if (isCompensated) {
        return res.send({ success: false, err: 'User has already been compensated'});
    }
    const compensationRecorded = recordCompensation(sessionId, email);
    if (!compensationRecorded) {
        return res.send({ success: false, err: 'Issue recording email. Please contact kleach@umich.edu'});
    }

    userData.comments = comments;

    const success = updateUser(userData);
    if (!success) {
        return res.send({ success: false, err: 'Could not save comments.' });
    }
    
    return res.send({ success: true });
});

router.post('/snippet/:section/:id', (req, res) => {
    const section = req.params.section;
    const snippetIdx = parseInt(req.params.id);
    const { fid, questionKey, answer, sessionId } = req.body;
    const userData = getUserData(sessionId);

    if (section != userData.sections[userData.sectionIdx]) {
        return res.send({ success: false });
    } else if (fid != userData.snippets[section][snippetIdx].fid) {
        return res.send({ success: false });
    }

    const data = userData.snippets[section];

    // Does answer already exist for update
    const response = userData.responses[section][fid];
    if (response) {
        const questionResponse = _.find(response.questions, q => q.question == questionKey);
        if (!questionResponse) {
            return res.send({ success: false, err: 'Attempting to answer non-existant question'});
        }
        if (questionResponse.answer == undefined) {
            questionResponse.answer = [];
        }
        questionResponse.answer.push({ response: answer, time: Date.now() });
    } else {
        return res.send({ success: false, err: 'Attempting to answer non-existant section'});
    }

    const success = updateUser(userData);
    if (!success) {
        return res.send({ success: false });
    }

    return res.send({ success: true })
});

router.post('/code/:section/:id', (req, res) => {
    const section = req.params.section;
    const snippetIdx = parseInt(req.params.id);
    const { fid, questionKey, code, sessionId } = req.body;
    const userData = getUserData(sessionId);

    if (section != userData.sections[userData.sectionIdx]) {
        return res.send({ success: false });
    } else if (fid != userData.snippets[section][snippetIdx].fid) {
        return res.send({ success: false });
    }

    // Does answer already exist for update
    const response = userData.responses[section][fid];
    if (response) {
        const questionResponse = _.find(response.questions, q => q.question == questionKey);
        if (!questionResponse) {
            return res.send({ success: false, err: 'Attempting to answer non-existant question'});
        }
        const numResponses = _.get(questionResponse, 'answer.length') || 0;
        if (numResponses + 1 > MAX_CODE_SUBMISSIONS) {
            return res.send({ success: false, err: 'Exceeded maximum amount of attempts for this question.'});
        } 

        // Inject user response into class's code
        const groundTruth = userData.snippets[section][snippetIdx];
        const userCode = _.replace(groundTruth.snippet, '${userResponse}', code);

        // Submit to Yash's API
        // NOTE: not sure if class is the right key. Also you will need to add the "class" key to all snippets
        // in backend/data/editable_snippets ad backend/data/repair_snippets
        axios.post('http://127.0.0.1:5000/', { code: userCode, classname: groundTruth.class }).then(response => {
            const testCaseResults = _.map(response.data, o => {
                if (o.output === undefined) {
                    const did_pass = o.pass ? 'passed' : 'failed';
                    o.output = `Test case ${did_pass}.`;
                }
                return o;
            });
            if (questionResponse.answer === undefined) {
                questionResponse.answer = [];
            }
            questionResponse.answer.push({ results: testCaseResults, response: code, time: Date.now() });

            const success = updateUser(userData);
            if (!success) {
                return res.send({ success: false, attempts: [numResponses+1, MAX_CODE_SUBMISSIONS] });
            }
            return res.send({ success: true, data: testCaseResults, attempts: [numResponses+1, MAX_CODE_SUBMISSIONS] });
        });
    } else {
        return res.send({ success: false, err: 'Attempting to answer non-existant section'});
    }
});

router.post('/nextQuestion', (req, res) => {
    const { sessionId } = req.body;
    const userData = getUserData(sessionId);
    const { sectionIdx, snippetIdx, sections, snippets, responses, questionIdx } = userData;
    const maxSnippetValues = { 'comprehension': 10, 'methodWrite': 1, 'repair': 1 };
    
    let relevantSection = sections[sectionIdx];
    if (questionIdx + 1 < snippets[relevantSection][snippetIdx].questions.length) {
        userData.questionIdx += 1;
    } else if (snippetIdx + 1 < maxSnippetValues[relevantSection]) {
        userData.snippetIdx += 1;
        userData.questionIdx = 0;
    } else {
        // move on to next section
        userData.snippetIdx = 0;
        userData.questionIdx = 0;
        userData.sectionIdx += 1;
        if (userData.sectionIdx >= sections.length) {
            userData.completed = true;
            const completed = markCompleted(sessionId, req.headers['x-forwarded-for']);
            userData.surveyEnd = Date.now();
            if (!completed) {
                console.log('not completed');
                return res.send({ success: false });
            }
        } else {
            relevantSection = sections[userData.sectionIdx];
        }
    }

    if (userData.completed && updateUser(userData)) {
        return res.send({ success: true, section: 'completed' });
    }

    const relevantSnippets = snippets[relevantSection];
    const fid = relevantSnippets[userData.snippetIdx].fid;
    const question = relevantSnippets[userData.snippetIdx].questions[userData.questionIdx];
    const sectionResponse = responses[relevantSection][fid];
    const questionResponse = _.find(_.get(sectionResponse, 'questions'), o => o.question == question.data);
    if (!sectionResponse) {
        responses[relevantSection][fid] = {
            start: Date.now(),
            questions: []
        }
    }
    if (!questionResponse) {
        responses[relevantSection][fid].questions.push({ question: question.data, start: Date.now() });
    }

    const success = updateUser(userData);
    if (!success) {
        return res.send({ success: false });
    }

    return res.send({
        success: true,
        data: relevantSnippets[userData.snippetIdx],
        snippetIdx: userData.snippetIdx,
        section: relevantSection,
        question,
        response: questionResponse
    });
});

module.exports = router;